#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <string.h>
#include <linux/videodev2.h>

#define ZERO(a) (a); memset(&(a), 0, sizeof(a));
#define test_ioctl(fd, request, ...) do {		\
	if(0 != ioctl(fd, request, __VA_ARGS__)) {	\
		perror("ioctl error: "#request);	\
		exit(1);				\
	}						\
} while(0);

static int fd;

int test_open(const char *devname)
{
	int fd = open(devname, O_RDWR);
	if(fd < 0) {
		perror("open");
		exit(1);
	}
	return fd;
}

void test_close(int fd)
{
	if(close(fd) != 0) {
		perror("close");
		exit(1);
	}
}

void *test_mmap(int fd, size_t length, off_t offset)
{
	void *addr = mmap(NULL, length, PROT_READ | PROT_WRITE, MAP_SHARED,
				fd, offset);
	if(addr == MAP_FAILED) {
		perror("mmap");
		exit(1);
	}
	return addr;
}

const char* test_type(enum v4l2_buf_type type)
{
	switch(type) {
		case V4L2_BUF_TYPE_VIDEO_CAPTURE: return "CAP:";
		case V4L2_BUF_TYPE_VIDEO_OUTPUT : return "OUT:";
		default: break;
	}
	return "";
}

void test_format(int w, int h, int pf, enum v4l2_buf_type type)
{
	struct v4l2_format ZERO(f);
	f.type = type;
	test_ioctl(fd, VIDIOC_G_FMT, &f);
	printf("%s format get: %08x\n", test_type(type),
					f.fmt.pix.pixelformat);

	f.fmt.pix.width  = w;
	f.fmt.pix.height = h;
	f.fmt.pix.pixelformat = pf;
	test_ioctl(fd, VIDIOC_S_FMT, &f);
	printf("%s format set: %08x (%dx%d) -> (%dx%d)\n", test_type(type),
		f.fmt.pix.pixelformat, w, h,
		f.fmt.pix.width, f.fmt.pix.height);
}

void test_rbufs(int n, enum v4l2_buf_type type)
{
	struct v4l2_requestbuffers ZERO(rbuf);
	rbuf.count  = n;
	rbuf.type   = type;
	rbuf.memory = V4L2_MEMORY_MMAP;
	test_ioctl(fd, VIDIOC_REQBUFS, &rbuf);
	printf("%s rbufs: %d\n", test_type(type), rbuf.count);
}

void test_qbufs(int i, int type)
{
	struct v4l2_buffer ZERO(qbuf);
	qbuf.index  = i;
	qbuf.type   = type;
	qbuf.memory = V4L2_MEMORY_MMAP;
	test_ioctl(fd, VIDIOC_QUERYBUF, &qbuf);

	void *addr = test_mmap(fd, qbuf.length, qbuf.m.offset);
	printf("%s qbufs: %2d %08x(%8x)\n", test_type(type),
		qbuf.index, qbuf.m.offset, qbuf.length);
	
	if(qbuf.type == V4L2_BUF_TYPE_VIDEO_OUTPUT) {
		qbuf.bytesused = qbuf.length;
		qbuf.timestamp.tv_sec = 0;
		qbuf.timestamp.tv_usec = 0;
		memset(addr, 0x80 + i, qbuf.length);
	}
	else
		memset(addr, 0x10 + i, qbuf.length);
	test_ioctl(fd, VIDIOC_QBUF, &qbuf);
	munmap(addr, qbuf.length);
}

void test_stream(void)
{
	enum v4l2_buf_type cap = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	enum v4l2_buf_type out = V4L2_BUF_TYPE_VIDEO_OUTPUT;
	test_ioctl(fd, VIDIOC_STREAMON, &cap);
	test_ioctl(fd, VIDIOC_STREAMON, &out);

	fd_set rfds;
	fd_set wfds;
	struct timeval tv;
	int ret;

	FD_ZERO(&rfds);
	FD_ZERO(&wfds);
	FD_SET(fd, &rfds);
	FD_SET(fd, &wfds);
	tv.tv_sec  = 2;
	tv.tv_usec = 0;
	ret = select(fd + 1, &rfds, &wfds, NULL, &tv);
	if(ret < 0)
		perror("select error");
	else if(ret == 0)
		perror("select timeout");
	else {
		const char *r = FD_ISSET(fd, &rfds) ? "r" : " ";
		const char *w = FD_ISSET(fd, &wfds) ? "w" : " ";
		printf("select: %s%s\n", r, w);
	}
}

void test_dqbufs(int type)
{
	struct v4l2_buffer ZERO(dqbuf);
	dqbuf.type   = type;
	dqbuf.memory = V4L2_MEMORY_MMAP;

	test_ioctl(fd, VIDIOC_DQBUF, &dqbuf);
	unsigned int prevoffset = dqbuf.m.offset;
/*
	this is odd,
	for the capture type, DQBUF is returning the wrong buffer offset.
	QUERYBUF must be called to correct this offset.
*/
	test_ioctl(fd, VIDIOC_QUERYBUF, &dqbuf);

	unsigned char *addr = test_mmap(fd, dqbuf.length, dqbuf.m.offset);
	printf("%s dqbufs: %2d (%08x|%08x) %02x%02x %02x %02x\n",
		test_type(type),
		dqbuf.index, prevoffset, dqbuf.m.offset,
		addr[0], addr[1], addr[2], addr[3]);
	munmap(addr, dqbuf.length);
}

int main(int argc, char *argv[])
{
	fd = test_open("/dev/video2");

	struct v4l2_capability ZERO(cap);
	test_ioctl(fd, VIDIOC_QUERYCAP, &cap);
	printf("querycap: %s/%s (%s)\n", cap.driver, cap.card, cap.bus_info);

	test_format(642, 935, V4L2_PIX_FMT_RGB565, V4L2_BUF_TYPE_VIDEO_CAPTURE);
	test_format(890, 485, V4L2_PIX_FMT_RGB565, V4L2_BUF_TYPE_VIDEO_OUTPUT);

	test_rbufs(10, V4L2_BUF_TYPE_VIDEO_CAPTURE);
	test_rbufs(10, V4L2_BUF_TYPE_VIDEO_OUTPUT);


	test_qbufs(0, V4L2_BUF_TYPE_VIDEO_CAPTURE);
	test_qbufs(1, V4L2_BUF_TYPE_VIDEO_OUTPUT);
	test_qbufs(5, V4L2_BUF_TYPE_VIDEO_CAPTURE);
	test_qbufs(7, V4L2_BUF_TYPE_VIDEO_OUTPUT);

	test_stream();

	test_dqbufs(V4L2_BUF_TYPE_VIDEO_CAPTURE);
	test_dqbufs(V4L2_BUF_TYPE_VIDEO_OUTPUT);
	test_dqbufs(V4L2_BUF_TYPE_VIDEO_CAPTURE);
	test_dqbufs(V4L2_BUF_TYPE_VIDEO_OUTPUT);

	test_close(fd);
	return 0;
}

